/* ----------------------------------------------------------------------------
 * Hotkey.cpp --
 *
 *	Wrapper to Win32's RegisterHotKey() done as an [Incr Tcl] class
 *	extension that happens to be written in C++ for that warm and cushy
 *	OO feeling.
 *
 * ----------------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this.  See license.txt
 * for more detail.
 *
 * ----------------------------------------------------------------------------
 * RCS: @(#) $Id: Hotkey.cpp,v 1.12 2003/04/06 23:27:49 davygrvy Exp $
 * ----------------------------------------------------------------------------
 */

#ifdef _MSC_VER
#   pragma comment (lib, "user32.lib")
#endif

#include "Mcl/include/CMcl.h"

// for the namespace and callframe APIs
#include "tclInt.h"

#include "cpptcl/ItclAdaptor.hpp"
#include "cpptcl/TclHash.hpp"

// ----------------------------------------------------------------------
// Platform SDK specifics that stock VC6 knows nothing about.
// ----------------------------------------------------------------------
#ifndef VK_SLEEP
#   define VK_SLEEP		0x5F
#endif
#ifndef VK_BROWSER_BACK
#   define VK_BROWSER_BACK	0xA6
#endif
#ifndef VK_BROWSER_FORWARD
#   define VK_BROWSER_FORWARD	0xA7
#endif
#ifndef VK_BROWSER_REFRESH
#   define VK_BROWSER_REFRESH	0xA8
#endif
#ifndef VK_BROWSER_STOP
#   define VK_BROWSER_STOP	0xA9
#endif
#ifndef VK_BROWSER_SEARCH
#   define VK_BROWSER_SEARCH	0xAA
#endif
#ifndef VK_BROWSER_FAVORITES
#   define VK_BROWSER_FAVORITES	0xAB
#endif
#ifndef VK_BROWSER_HOME
#   define VK_BROWSER_HOME	0xAC
#endif
#ifndef VK_VOLUME_MUTE
#   define VK_VOLUME_MUTE	0xAD
#endif
#ifndef VK_VOLUME_DOWN
#   define VK_VOLUME_DOWN	0xAE
#endif
#ifndef VK_VOLUME_UP
#   define VK_VOLUME_UP		0xAF
#endif
#ifndef VK_MEDIA_NEXT_TRACK
#   define VK_MEDIA_NEXT_TRACK	0xB0
#endif
#ifndef VK_MEDIA_PREV_TRACK
#   define VK_MEDIA_PREV_TRACK	0xB1
#endif
#ifndef VK_MEDIA_STOP
#   define VK_MEDIA_STOP	0xB2
#endif
#ifndef VK_MEDIA_PLAY_PAUSE
#   define VK_MEDIA_PLAY_PAUSE	0xB3
#endif
#ifndef VK_LAUNCH_MAIL
#   define VK_LAUNCH_MAIL	0xB4
#endif
#ifndef VK_LAUNCH_MEDIA_SELECT
#   define VK_LAUNCH_MEDIA_SELECT 0xB5
#endif
#ifndef VK_LAUNCH_APP1
#   define VK_LAUNCH_APP1	0xB6
#endif
#ifndef VK_LAUNCH_APP2
#   define VK_LAUNCH_APP2	0xB7
#endif
#ifndef VK_OEM_1
#   define VK_OEM_1		0xBA	// ';:' for US
#endif
#ifndef VK_OEM_PLUS
#   define VK_OEM_PLUS		0xBB	// '+' any country
#endif
#ifndef VK_OEM_COMMA
#   define VK_OEM_COMMA		0xBC	// ',' any country
#endif
#ifndef VK_OEM_MINUS
#   define VK_OEM_MINUS		0xBD	// '-' any country
#endif
#ifndef VK_OEM_PERIOD
#   define VK_OEM_PERIOD	0xBE	// '.' any country
#endif
#ifndef VK_OEM_2
#   define VK_OEM_2		0xBF	// '/?' for US
#endif
#ifndef VK_OEM_3
#   define VK_OEM_3		0xC0	// '`~' for US
#endif
#ifndef VK_OEM_4
#   define VK_OEM_4		0xDB	//  '[{' for US
#endif
#ifndef VK_OEM_5
#   define VK_OEM_5		0xDC	//  '\|' for US
#endif
#ifndef VK_OEM_6
#   define VK_OEM_6		0xDD	//  ']}' for US
#endif
#ifndef VK_OEM_7
#   define VK_OEM_7		0xDE	//  ''"' for US
#endif
#ifndef VK_OEM_8
#   define VK_OEM_8		0xDF
#endif
#ifndef VK_OEM_AX
#   define VK_OEM_AX		0xE1	//  'AX' key on Japanese AX kbd
#endif
#ifndef VK_OEM_102
#   define VK_OEM_102		0xE2	//  '<>' or '\|' on RT 102-key kbd.
#endif
#ifndef VK_ICO_HELP
#   define VK_ICO_HELP		0xE3	//  Help key on ICO
#endif
#ifndef VK_ICO_00
#   define VK_ICO_00		0xE4	//  00 key on ICO
#endif
#ifndef VK_OEM_RESET
#   define VK_OEM_RESET		0xE9	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_JUMP
#   define VK_OEM_JUMP		0xEA	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_PA1
#   define VK_OEM_PA1		0xEB	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_PA2
#   define VK_OEM_PA2		0xEC	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_PA3
#   define VK_OEM_PA3		0xED	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_WSCTRL
#   define VK_OEM_WSCTRL	0xEE	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_CUSEL
#   define VK_OEM_CUSEL		0xEF	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_ATTN
#   define VK_OEM_ATTN		0xF0	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_FINISH
#   define VK_OEM_FINISH	0xF1	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_COPY
#   define VK_OEM_COPY		0xF2	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_AUTO
#   define VK_OEM_AUTO		0xF3	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_ENLW
#   define VK_OEM_ENLW		0xF4	//  Nokia/Ericsson definitions
#endif
#ifndef VK_OEM_BACKTAB
#   define VK_OEM_BACKTAB	0xF5	//  Nokia/Ericsson definitions
#endif


// ----------------------------------------------------------------------
// Format system error messages.
// ----------------------------------------------------------------------
static Tcl_Obj *
GetSysMsgObj (DWORD id)
{
    int chars;
    char sysMsgSpace[512];

    chars = wsprintf(sysMsgSpace, "[%u] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_IGNORE_INSERTS |
	    FORMAT_MESSAGE_MAX_WIDTH_MASK, 0L, id, 0, sysMsgSpace, (512-chars),
	    0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}


// ----------------------------------------------------------------------
// Virtual key table.  Hide the Tcl::Hash class so we enforce read-only.
// ----------------------------------------------------------------------
class VirtKeyHash : private Tcl::Hash<UINT>
{
public:
    VirtKeyHash()
    {
	Add("break",	    VK_CANCEL);
	Add("backspace",    VK_BACK);
	Add("tab",	    VK_TAB);
	Add("clear",	    VK_CLEAR);
	Add("return",	    VK_RETURN);
	Add("pause",	    VK_PAUSE);
	Add("capslock",	    VK_CAPITAL);
	Add("ime:kana",	    VK_KANA);
	Add("ime:hangul",   VK_HANGUL);
	Add("ime:junja",    VK_JUNJA);
	Add("ime:final",    VK_FINAL);
	Add("ime:hanja",    VK_HANJA);
	Add("ime:kanji",    VK_KANJI);
	Add("escape",	    VK_ESCAPE);
	Add("ime:convert",  VK_CONVERT);
	Add("ime:nonconvert", VK_NONCONVERT);
	Add("ime:accept",   VK_ACCEPT);
	Add("ime:mode",	    VK_MODECHANGE);
	Add("spacebar",	    VK_SPACE);
	Add("prior",	    VK_PRIOR);
	Add("next",	    VK_NEXT);
	Add("end",	    VK_END);
	Add("home",	    VK_HOME);
	Add("left",	    VK_LEFT);
	Add("up",	    VK_UP);
	Add("right",	    VK_RIGHT);
	Add("down",	    VK_DOWN);
	Add("select",	    VK_SELECT);
	Add("print",	    VK_PRINT);
	Add("execute",	    VK_EXECUTE);
	Add("prntscr",	    VK_SNAPSHOT);
	Add("insert",	    VK_INSERT);
	Add("delete",	    VK_DELETE);
	Add("help",	    VK_HELP);
	Add("0",	    '0');
	Add("1",	    '1');
	Add("2",	    '2');
	Add("3",	    '3');
	Add("4",	    '4');
	Add("5",	    '5');
	Add("6",	    '6');
	Add("7",	    '7');
	Add("8",	    '8');
	Add("9",	    '9');
	Add("A",	    'A');
	Add("B",	    'B');
	Add("C",	    'C');
	Add("D",	    'D');
	Add("E",	    'E');
	Add("F",	    'F');
	Add("G",	    'G');
	Add("H",	    'H');
	Add("I",	    'I');
	Add("J",	    'J');
	Add("K",	    'K');
	Add("L",	    'L');
	Add("M",	    'M');
	Add("N",	    'N');
	Add("O",	    'O');
	Add("P",	    'P');
	Add("Q",	    'Q');
	Add("R",	    'R');
	Add("S",	    'S');
	Add("T",	    'T');
	Add("U",	    'U');
	Add("V",	    'V');
	Add("W",	    'W');
	Add("X",	    'X');
	Add("Y",	    'Y');
	Add("Z",	    'Z');
	Add("sleep",	    VK_SLEEP);
	Add("num:0",	    VK_NUMPAD0);
	Add("num:1",	    VK_NUMPAD1);
	Add("num:2",	    VK_NUMPAD2);
	Add("num:3",	    VK_NUMPAD3);
	Add("num:4",	    VK_NUMPAD4);
	Add("num:5",	    VK_NUMPAD5);
	Add("num:6",	    VK_NUMPAD6);
	Add("num:7",	    VK_NUMPAD7);
	Add("num:8",	    VK_NUMPAD8);
	Add("num:9",	    VK_NUMPAD9);
	Add("num:*",	    VK_MULTIPLY);
	Add("num:plus",	    VK_ADD);
//	Add("num:enter",    VK_SEPARATOR);
	Add("num:-",	    VK_SUBTRACT);
	Add("num:.",	    VK_DECIMAL);
	Add("num:/",	    VK_DIVIDE);
	Add("f1",	    VK_F1);
	Add("f2",	    VK_F2);
	Add("f3",	    VK_F3);
	Add("f4",	    VK_F4);
	Add("f5",	    VK_F5);
	Add("f6",	    VK_F6);
	Add("f7",	    VK_F7);
	Add("f8",	    VK_F8);
	Add("f9",	    VK_F9);
	Add("f10",	    VK_F10);
	Add("f11",	    VK_F11);
	Add("f12",	    VK_F12);
	Add("f13",	    VK_F13);
	Add("f14",	    VK_F14);
	Add("f15",	    VK_F15);
	Add("f16",	    VK_F16);
	Add("f17",	    VK_F17);
	Add("f18",	    VK_F18);
	Add("f19",	    VK_F19);
	Add("f20",	    VK_F20);
	Add("f21",	    VK_F21);
	Add("f22",	    VK_F22);
	Add("f23",	    VK_F23);
	Add("f24",	    VK_F24);
	Add("numlock",	    VK_NUMLOCK);
	Add("scrollock",    VK_SCROLL);
	Add("browse:back",  VK_BROWSER_BACK);
	Add("browse:forward", VK_BROWSER_FORWARD);
	Add("browse:refresh", VK_BROWSER_REFRESH);
	Add("browse:stop",  VK_BROWSER_STOP);
	Add("browse:search", VK_BROWSER_SEARCH);
	Add("browse:fav",   VK_BROWSER_FAVORITES);
	Add("browse:home",  VK_BROWSER_HOME);
	Add("volume:mute",  VK_VOLUME_MUTE);
	Add("volume:down",  VK_VOLUME_DOWN);
	Add("volume:up",    VK_VOLUME_UP);
	Add("media:next",   VK_MEDIA_NEXT_TRACK);
	Add("media:prev",   VK_MEDIA_PREV_TRACK);
	Add("media:stop",   VK_MEDIA_STOP);
	Add("media:play",   VK_MEDIA_PLAY_PAUSE);
	Add("launch:mail",  VK_LAUNCH_MAIL);
	Add("launch:select", VK_LAUNCH_MEDIA_SELECT);
	Add("launch:app1",  VK_LAUNCH_APP1);
	Add("launch:app2",  VK_LAUNCH_APP2);
	Add("oem:1",	    VK_OEM_1);		//  ';:' for US
	Add("oem:plus",	    VK_OEM_PLUS);	//  '+' any country
	Add("oem:,",	    VK_OEM_COMMA);	//  ',' any country
	Add("oem:-",	    VK_OEM_MINUS);	//  '-' any country
	Add("oem:.",	    VK_OEM_PERIOD);	//  '.' any country
	Add("oem:2",	    VK_OEM_2);		//  '/?' for US
	Add("oem:3",	    VK_OEM_3);		//  '`~' for US
	Add("oem:4",	    VK_OEM_4);		//  '[{' for US
	Add("oem:5",	    VK_OEM_5);		//  '\|' for US
	Add("oem:6",	    VK_OEM_6);		//  ']}' for US
	Add("oem:7",	    VK_OEM_7);		//  ''"' for US
	Add("oem:8",	    VK_OEM_8);		//  unknown!
	Add("ax",	    VK_OEM_AX);		//  'AX' key on Japanese AX kbd
	Add("oem:102",	    VK_OEM_102);	//  '<>' or '\|' on RT 102-key kbd.
	Add("ico:help",	    VK_ICO_HELP);	//  Help key on ICO
	Add("ico:oo",	    VK_ICO_00);		//  00 key on ICO
	Add("ime:process",  VK_PROCESSKEY);
	Add("nokia:reset",  VK_OEM_RESET);
	Add("nokia:jump",   VK_OEM_JUMP);
	Add("nokia:pa1",    VK_OEM_PA1);
	Add("nokia:pa2",    VK_OEM_PA2);
	Add("nokia:pa3",    VK_OEM_PA3);
	Add("nokia:wsctrl", VK_OEM_WSCTRL);
	Add("nokia:cusel",  VK_OEM_CUSEL);
	Add("nokia:attn",   VK_OEM_ATTN);
	Add("nokia:finish", VK_OEM_FINISH);
	Add("nokia:copy",   VK_OEM_COPY);
	Add("nokia:auto",   VK_OEM_AUTO);
	Add("nokia:enlw",   VK_OEM_ENLW);
	Add("nokia:backtab", VK_OEM_BACKTAB);
	Add("attn",	    VK_ATTN);
	Add("crsel",	    VK_CRSEL);
	Add("exsel",	    VK_EXSEL);
	Add("ereof",	    VK_EREOF);
	Add("play",	    VK_PLAY);
	Add("zoom",	    VK_ZOOM);
	Add("reserved",	    VK_NONAME);
	Add("pa1",	    VK_PA1);
	Add("oem:clear",    VK_OEM_CLEAR);
    }
    int Find (const char *key, UINT *valuePtr)
    {
	// C++ compiler bug!  This member function is not static,
	// so don't give me any lip.  Should be:
	// return Tcl::Hash<UINT>::Find(key, valuePtr);
	return ((Tcl::Hash<UINT>*)this)->Find(key, valuePtr);
    }
};


// ----------------------------------------------------------------------
// This is the worker thread class for setting up and listening for hotkey
// announcements.  Note that from the thread, the Tcl_Obj of the script is
// never modified or manipulated in any way.
// ----------------------------------------------------------------------
class HotKeyListener : public CMclThreadHandler
{
public:
    HotKeyListener(UINT _vk, UINT _fsModifiers, Tcl_Obj *_script,
	CMclQueue<Tcl_Obj *> &_workQ, Tcl_AsyncHandler _Async, CMclEvent &_gotReg)
	: vk(_vk), fsModifiers(_fsModifiers), workQ(_workQ), Async(_Async),
	script(_script), gotReg(_gotReg), status(NO_ERROR)
    {
	// take ownership.
	Tcl_IncrRefCount(script);
    }
    ~HotKeyListener()
    {
	// release ownership.
	Tcl_DecrRefCount(script);
    }
    DWORD Status ()
    {
	return status;
    }
private:
    // This is the routine run within a thread.
    //
    unsigned ThreadHandlerProc(void)
    {
	MSG msg;
	BOOL ok;

	ok = RegisterHotKey(0L, 1, fsModifiers, vk);
	if (!ok) {
	    status = GetLastError();
	    gotReg.Set();
	    return status;
	}

	// Let the main thread know we're up without an error.
	gotReg.Set();

	while (GetMessage(&msg, 0L, 0, 0) > 0) {
	    switch (msg.message) {
	    case WM_HOTKEY:
		// I'll be extra cautious and use SEH because I trust
		// no one ;)
		__try {
		    workQ.Put(script);
		    Tcl_AsyncMark(Async);
		}
		__except (EXCEPTION_EXECUTE_HANDLER) {
		    UnregisterHotKey(0L, 1);
		    return _exception_code();
		}
	    }
	}

	UnregisterHotKey(0L, 1);

	// By convention, the exit code from the WM_QUIT is used as the
	// return value.
	//
	return msg.wParam;
    }
    UINT vk, fsModifiers;
    CMclQueue<Tcl_Obj *> &workQ;
    Tcl_AsyncHandler Async;
    Tcl_Obj *script;
    CMclEvent &gotReg;
    DWORD status;
};


// ----------------------------------------------------------------------
// This class manages the Itcl interface.
// ----------------------------------------------------------------------
class HotkeyItclClassExtension : protected Itcl::IAdaptor<HotkeyItclClassExtension>
{
    // allows DoCleanup() to be private.
    friend class Tcl::Adaptor<HotkeyItclClassExtension>;

    // This structure is what gets queued into Tcl's event loop
    struct HOTKEYEVENT {
	Tcl_Event header;
	Tcl_Obj *script;
	HotkeyItclClassExtension *ext;
    };

    // We store this in the hash table
    struct HOTKEYINFO {
      CMclThread *thread;
      HotKeyListener *listener;
    };

    Tcl::Hash<HOTKEYINFO *, TCL_ONE_WORD_KEYS> hkHash;
				    // All our hotkey listeners.

    VirtKeyHash virtKeyTable;	    // Read-only lookup for virtual key
				    // codes.

    Tcl_AsyncHandler Async;	    // Interp specific async handler.
    CMclQueue<Tcl_Obj *> workQ;	    // The work queue for jobs.

public:
    HotkeyItclClassExtension(Tcl_Interp *interp)
	: Itcl::IAdaptor<HotkeyItclClassExtension>(interp)
    {
	NewItclCmd("winutils-hotkey-construct", HotKeyConstructCmd);
	NewItclCmd("winutils-hotkey-destruct", HotKeyDestructCmd);
	Async = Tcl_AsyncCreate(FlushHotkeys, this);
    }
    virtual ~HotkeyItclClassExtension()
    {
	// We are NOT allowed to use the interp pointer in here.
	//
	HOTKEYINFO *hkInfo;
	int result;

        for (
	    result = hkHash.Top(&hkInfo);
	    result == TCL_OK;
	    result = hkHash.Next(&hkInfo)
        ) {
	    // Cause the message pump to shutdown.
	    //
	    PostThreadMessage(hkInfo->thread->GetThreadId(), WM_QUIT, 0, 0);

	    // The thread handle is placed in the signaled state when it
	    // finishes.  Use a timeout just to be safe to avoid a lock-up.
	    //
	    hkInfo->thread->Wait(500);

	    delete hkInfo->thread;
	    delete hkInfo->listener;
	    delete hkInfo;
	}

	// Now that all the HotKeyListener instances that had a reference to
	// the Async token are closed, close the Async token.
	//
	Tcl_AsyncDelete(Async);

	// Now remove any events currently in the event loop that call
	// into us.  This is very important because the Tcl_Obj* for the
	// script within the Tcl_Event structure (really a HOTKEYEVENT)
	// is now a dangling reference and most likely had been returned
	// to the allocator since the HotKeyListener instance gave up
	// ownership to the Tcl_Obj* when it was closed above.
	//
	Tcl_DeleteEvents(HoseHotkeyEvents, this);
    }

private:

    // ------------------------------------------------------------------
    // This is called by the adapter base to let us know we are about to
    // go away and it's safe to use the interp for any needed cleanup.
    // The destructor is called immediatly after this.
    // ------------------------------------------------------------------
    virtual void DoCleanup() {}

    // ------------------------------------------------------------------
    // Constructor method
    // ------------------------------------------------------------------
    int HotKeyConstructCmd (int objc, struct Tcl_Obj * const objv[])
    {
	HOTKEYINFO *hkInfo;
	CMclEvent gotReg;
	DWORD err;
	UINT vk, fsModifiers = 0;
	Tcl_Obj *listPtr, **hkList;
	int i, hkParts, index;
	const char *vKey;
	static const char *modifiers[] = {"Alt", "Shift", "Ctrl", "Win", 0L};
	enum modifier {HOTKEY_ALT, HOTKEY_SHIFT, HOTKEY_CTRL, HOTKEY_WIN};
	ItclObject *HotkeyItclObj;

	if (objc != 3) {
	    Tcl_WrongNumArgs(interp, 1, objv, "<keyString> <script>");
	    return TCL_ERROR;
	}

	// Get our Itcl instance pointer.
	//
	if (GetItclObj(&HotkeyItclObj, objv[0]) != TCL_OK) return TCL_ERROR;

	SplitHotKeyParts(objv[1], &listPtr);
	Tcl_ListObjGetElements(0L, listPtr, &hkParts, &hkList);

	if (hkParts > 1) {
	    for (i = 0; i < (hkParts - 1); i++) {
		if (Tcl_GetIndexFromObj(interp, hkList[i], modifiers, "modifier",
			TCL_EXACT, &index) != TCL_OK) {
		    Tcl_DecrRefCount(listPtr);
		    return TCL_ERROR;
		}
		switch ((enum modifier) index) {
		case HOTKEY_ALT:
		    fsModifiers |= MOD_ALT; break;
		case HOTKEY_SHIFT:
		    fsModifiers |= MOD_SHIFT; break;
		case HOTKEY_CTRL:
		    fsModifiers |= MOD_CONTROL; break;
		case HOTKEY_WIN:
		    fsModifiers |= MOD_WIN;
		}
	    }
	}

	// The last one is our virtual key name.
	//
	vKey = Tcl_GetString(hkList[hkParts - 1]);
	if (virtKeyTable.Find(vKey, &vk) != TCL_OK) {
	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		    "virtual key called \"", vKey, "\" is unknown", 0L);
	    Tcl_DecrRefCount(listPtr);
	    return TCL_ERROR;
	}
	Tcl_DecrRefCount(listPtr);

	hkInfo = new HOTKEYINFO;
	hkInfo->listener = new HotKeyListener(vk, fsModifiers,
		objv[2] /* script */, workQ, Async, gotReg);

	// The new thread is created here.  Use a tiny stack size.
	//
	try {
	    hkInfo->thread = new CMclThread(hkInfo->listener, 0, 0L, 256);
	}
	catch (DWORD err) {
       	    Tcl_SetObjResult(interp, GetSysMsgObj(err));
	    return TCL_ERROR;
	}

	// Block until after it calls RegisterHotkey() so we can know if
	// it had failed.
	//
	gotReg.Wait(INFINITE);
	if ((err = hkInfo->listener->Status()) != NO_ERROR) {
	    Tcl_SetObjResult(interp, GetSysMsgObj(err));
	    delete hkInfo->thread;
	    delete hkInfo->listener;
	    delete hkInfo;
	    return TCL_ERROR;
	}

	// save the new HOTKEYINFO structure into the hash table associated
	// to the Itcl instance.
	//
	hkHash.Add(HotkeyItclObj, hkInfo);
	return TCL_OK;
    }

    // ------------------------------------------------------------------
    // Destructor method
    // ------------------------------------------------------------------
    int HotKeyDestructCmd (int objc, struct Tcl_Obj * const objv[])
    {
	ItclObject *HotkeyItclObj;
	HOTKEYINFO *hkInfo;

	if (GetItclObj(&HotkeyItclObj, objv[0]) != TCL_OK) return TCL_ERROR;

	// Itcl's destructor may be called even though the Itcl class
	// constructor returned with an error!  Thus, no winutils::hotkey
	// instance exists in the hash table even though an Itcl object
	// context exists.  Only delete what we know is there.

	if (hkHash.Extract(HotkeyItclObj, &hkInfo) == TCL_OK) {

	    // Cause the message pump to shutdown.
	    //
	    PostThreadMessage(hkInfo->thread->GetThreadId(), WM_QUIT, 0, 0);

	    // The thread handle is placed in the signaled state when it
	    // finishes.  Use a timeout just to be extra safe to avoid a
	    // lock-up.
	    //
	    hkInfo->thread->Wait(500);

	    delete hkInfo->thread;
	    delete hkInfo->listener;
	    delete hkInfo;
	}
	return TCL_OK;
    }

    // ------------------------------------------------------------------
    // Split it into a list using '+' as the seperator.
    // ------------------------------------------------------------------
    void SplitHotKeyParts (Tcl_Obj *in, Tcl_Obj **listPtrPtr)
    {
	char *element, *string, *end;
	int stringLen, len;
	Tcl_UniChar ch;
	Tcl_Obj *objPtr;

	string = Tcl_GetStringFromObj(in, &stringLen);
	end = string + stringLen;
	*listPtrPtr = Tcl_NewObj();

	for (element = string; string < end; string += len) {
	    len = Tcl_UtfToUniChar(string, &ch);
	    if (ch == L'+') {
		objPtr = Tcl_NewStringObj(element, string - element);
		Tcl_ListObjAppendElement(0L, *listPtrPtr, objPtr);
		element = string + len;
	    }
	}
	objPtr = Tcl_NewStringObj(element, string - element);
	Tcl_ListObjAppendElement(0L, *listPtrPtr, objPtr);
    }

    // ------------------------------------------------------------------
    // Our Tcl_AsyncProc is calling back into us.
    // ------------------------------------------------------------------
    void QueueAllPendingHotkeys ()
    {
	Tcl_Obj *script;
	HOTKEYEVENT *ev;

	// WARNING: Only Tcl allocation routines are allowed in here.

	// Some running HotKeyListener instance called Tcl_AsyncMark.  There
	// must be at least one entry in the WorkQueue, but pull them all off
	// while we're here.  Don't let a good context go to waste.
	//
	while (workQ.Get(script, 0)) {

	    ev = reinterpret_cast<HOTKEYEVENT *>(
		    ckalloc(sizeof(HOTKEYEVENT)));
	    ev->header.proc = EvalCallback;
	    ev->ext    = this;	    // Used so EvalCallback knows which
				    // instance.
	    ev->script = script;    // you came from a distant thread

	    // Queue it into Tcl's event loop.
	    Tcl_QueueEvent(reinterpret_cast<Tcl_Event *>(ev),
		    TCL_QUEUE_TAIL);
	}
    }

    // ------------------------------------------------------------------
    // Our Tcl_EventProc is calling back into us from the event loop.
    // ------------------------------------------------------------------
    void EvalOneHotkey (Tcl_Obj *script)
    {
	CallFrame frame;
	Proc proc;
	Tcl_Namespace *nsPtr;
	int result;

	// let's have some fun and run it like a proc AND within a namespace
	nsPtr = Tcl_FindNamespace(interp, "::winutils", 0L, 0);
	Tcl_PushCallFrame(interp, (Tcl_CallFrame *)&frame, nsPtr, 1);

	proc.numCompiledLocals = 0;
	proc.firstLocalPtr = 0;
	proc.lastLocalPtr = 0;
	frame.procPtr = &proc;

	result = Tcl_EvalObj(interp, script);
	Tcl_PopCallFrame(interp);

	if (result == TCL_ERROR) {
	    Tcl_Obj *buff;
	    int len;
	    const char *str;
	    buff = Tcl_NewStringObj("\nError in hotkey script \"", -1);
	    Tcl_AppendObjToObj(buff, script);
	    Tcl_AppendStringsToObj(buff, "\"", 0L);
	    str = Tcl_GetStringFromObj(buff, &len);
	    Tcl_AddObjErrorInfo(interp, str, len);
	    Tcl_DecrRefCount(buff);
	    Tcl_BackgroundError(interp);
	}

	Tcl_ResetResult(interp);
    }

    // ------------------------------------------------------------------
    // The Tcl_AsyncProc callback.
    // ------------------------------------------------------------------
    static int FlushHotkeys (ClientData clientData, Tcl_Interp *, int code)
    {
	HotkeyItclClassExtension *ext =
		reinterpret_cast <HotkeyItclClassExtension *>(clientData);
	ext->QueueAllPendingHotkeys();
	return code;
    }

    // ------------------------------------------------------------------
    // The Tcl_EventProc callback.
    // ------------------------------------------------------------------
    static int EvalCallback (Tcl_Event *evPtr, int flags)
    {
	HOTKEYEVENT *event = reinterpret_cast<HOTKEYEVENT *>(evPtr);

	// we only handle file-type events here.
	if (!(flags & TCL_FILE_EVENTS)) return 0;
	(event->ext)->EvalOneHotkey(event->script);
	return 1;
    }

    // ------------------------------------------------------------------
    // The Tcl_EventDeleteProc callback.
    // ------------------------------------------------------------------
    static int HoseHotkeyEvents (Tcl_Event *evPtr, ClientData clientData)
    {
	// No dynamic storage was used in the structure, so we don't
	// need to do anything special in here.  return 1 if this
	// event belongs to us indicating that it should be deleted.
	//
	return (evPtr->proc == EvalCallback && 
	    (HotkeyItclClassExtension *)clientData == ((HOTKEYEVENT *)evPtr)->ext);
    }
};


// ----------------------------------------------------------------------
// The Tcl_PackageInitProc for this section.
// ----------------------------------------------------------------------
extern "C" int
Hotkey_Init (Tcl_Interp *interp)
{
#ifdef USE_ITCL_STUBS
    if (Itcl_InitStubs(interp, "3.1", 0) == 0L) {
	return TCL_ERROR;
    }
#else
    if (Itcl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
#endif
    new HotkeyItclClassExtension(interp);
    return TCL_OK;
}
