/* ----------------------------------------------------------------------------
 * RegisterService.c --
 *
 *	The reg98serv command.  Fully thread safe.  Code donated by
 *	
 *
 * ----------------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this.  See license.txt
 * for more detail.
 *
 * ----------------------------------------------------------------------------
 * RCS: @(#) $Id: RegisterService.c,v 1.3 2002/04/21 06:19:37 davygrvy Exp $
 * ----------------------------------------------------------------------------
 */

#include "tcl.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

static int initialized = 0;
static DWORD (WINAPI *lpfnRegSvcProc)(DWORD, DWORD) = 0L;
TCL_DECLARE_MUTEX(initLock)

static Tcl_Obj *GetSysMsgObj(DWORD id)
{
    int chars;
    char sysMsgSpace[512];

    chars = wsprintf(sysMsgSpace, "[%u] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
	    FORMAT_MESSAGE_IGNORE_INSERTS|FORMAT_MESSAGE_MAX_WIDTH_MASK,
	    0L, id, 0, sysMsgSpace+chars, (512-chars), 0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}


/* registerservice ?-process PID? true|false
 */
static int
RegServiceObjCmd (ClientData cdata, Tcl_Interp *interp,
                  int objc, Tcl_Obj *CONST objv[])
{
    Tcl_Obj *result = Tcl_GetObjResult(interp);
    int r = TCL_OK, cn = 1;
    DWORD dwType = 1;           /* register the service */
    DWORD dwProcessID = 0;      /* register the current process */

    if (objc > 4) {
        Tcl_WrongNumArgs(interp, 1, objv, "?-process PID? true|false");
        return TCL_ERROR;
    }

    if (objc > 2) {
	if (strncmp("-proc", Tcl_GetString(objv[cn]), 5) != 0) {
            const char msg[] = "invalid option: must be -process PID";
	    Tcl_SetObjResult(interp, Tcl_NewStringObj(msg, -1));
	    return TCL_ERROR;
	}
	cn++;
	r = Tcl_GetLongFromObj(interp, objv[cn], (long*)&dwProcessID);
	cn++;
    }

    if (r == TCL_OK && cn < objc) {
	r = Tcl_GetBooleanFromObj(interp, objv[cn], (int*)&dwType);
    }

    if (r == TCL_OK) {
	if (lpfnRegSvcProc != 0L) {
	    r = (lpfnRegSvcProc)(dwProcessID, dwType) ? TCL_OK : TCL_ERROR;
	    if (r != TCL_OK)
		Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	} else {
	    const char msg[] =
		    "error: RegisterServiceProcess() not available from OS.";
	    Tcl_SetObjResult(interp, Tcl_NewStringObj(msg, -1));
	    r = TCL_ERROR;
	}
    }
    return r;
}

int
Reg9XService_Init (Tcl_Interp *interp)
{
    Tcl_MutexLock(&initLock);
    if (!initialized) {
	HMODULE hKernel = GetModuleHandle("KERNEL32.DLL");
	
	initialized = 1;
	if (hKernel != 0L) {
	    (void *) lpfnRegSvcProc = GetProcAddress(hKernel,
		    "RegisterServiceProcess");
	}
    }
    Tcl_MutexUnlock(&initLock);
    Tcl_CreateObjCommand(interp, "winutils::reg9Xservice", RegServiceObjCmd, 0L, 0L);
    return TCL_OK;
}