/* ----------------------------------------------------------------------
 * SendMessage.cpp --
 *
 *	TODO: put something here.
 *
 * ----------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this code.  See
 * license.txt for more detail.
 *
 * ----------------------------------------------------------------------
 * RCS: @(#) $Id: SendMessage.cpp,v 1.4 2003/04/06 22:32:16 davygrvy Exp $
 * ----------------------------------------------------------------------
 */

#include "cpptcl/TclHash.hpp"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

TCL_DECLARE_MUTEX(initLock)
Tcl::Hash<UINT> *MsgTbl = 0L;


// ----------------------------------------------------------------------
// Format system error messages.
// ----------------------------------------------------------------------
static Tcl_Obj *
GetSysMsgObj (DWORD id)
{
    int chars;
    char sysMsgSpace[512];

    chars = wsprintf(sysMsgSpace, "[%u] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_IGNORE_INSERTS |
	    FORMAT_MESSAGE_MAX_WIDTH_MASK, 0L, id, 0, sysMsgSpace, (512-chars),
	    0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}

/*
 *----------------------------------------------------------------------
 *
 * CmdSendMessage --
 *
 *  uses the window's sendmessage function.
 *
 * Results:
 *  what sendmessage returns
 *
 * Side effects:
 *  we hope not.
 *
 *----------------------------------------------------------------------
 */
static int
SendMessageObjCmd (ClientData cData, Tcl_Interp *interp, int objc,
		   struct Tcl_Obj * CONST objv[])
{
    HWND hWnd = 0L;
    UINT Msg = 0;
    WPARAM wParam = 0L;
    LPARAM lParam = 0L;
    int num;
    DWORD result;
    LRESULT rtn;

    if (objc < 5 || objc > 5) {
	Tcl_WrongNumArgs(interp, 1, objv, "<hWnd> <msgnum|msgname> <wParam> <lParam>");
	return TCL_ERROR;
    }

    if (Tcl_GetIntFromObj(interp, objv[1], &num) != TCL_OK) {
	Tcl_AppendResult(interp, " <hWnd> didn't convert to a decimal.", NULL);
	return TCL_ERROR;
    }
    hWnd = (HWND) num;

    if (Tcl_GetIntFromObj(NULL, objv[2], &num) != TCL_OK) {
	// it isn't a number, so it must be the name of
	// a message.  Look it up in the message hash table.
	//
	if (MsgTbl->Find(Tcl_GetString(objv[2]), &Msg) == TCL_ERROR) {
	    // no dice.
	    Tcl_AppendResult(interp, "Message name (", Tcl_GetString(objv[2]), ") was not found in the list", NULL);
	    return TCL_ERROR;
	}
    } else {
	Msg = (UINT) num;
    }

    // we'll need some more stuff here for converting wParam and lParam
    if (Tcl_GetIntFromObj(interp, objv[3], &num) != TCL_OK) {
	Tcl_AppendResult(interp, " <wParam> didn't convert to a decimal.", 0L);
	return TCL_ERROR;
    }
    wParam = (WPARAM) num;

    if (Tcl_GetIntFromObj(interp, objv[4], &num) != TCL_OK) {
	Tcl_AppendResult(interp, " <lParam> didn't convert to a decimal.", 0L);
	return TCL_ERROR;
    }
    lParam = (LPARAM) num;

    // The call.
    rtn = SendMessageTimeout(hWnd, Msg, wParam, lParam, SMTO_ABORTIFHUNG,
	    4000, &result);

    if (rtn == 0) {
	DWORD err = GetLastError();
	if (err = NO_ERROR) {
	    Tcl_SetObjResult(interp, Tcl_NewStringObj(
		    "Operation timed-out.", -1));
	} else {
	    Tcl_SetObjResult(interp, GetSysMsgObj(err));
	}
	return TCL_ERROR;
    }

    Tcl_SetObjResult(interp, Tcl_NewLongObj(result));
    return TCL_OK;
};


void Fill_MsgTable ()
{
    MsgTbl = new Tcl::Hash<UINT>;
    MsgTbl->Add("WM_NULL", WM_NULL);
    MsgTbl->Add("WM_CREATE", WM_CREATE);
    MsgTbl->Add("WM_DESTROY", WM_DESTROY);
    MsgTbl->Add("WM_MOVE", WM_MOVE);
    MsgTbl->Add("WM_SIZE", WM_SIZE);
    MsgTbl->Add("WM_ACTIVATE", WM_ACTIVATE);
    MsgTbl->Add("WM_SETFOCUS", WM_SETFOCUS);
    MsgTbl->Add("WM_KILLFOCUS", WM_KILLFOCUS);
    MsgTbl->Add("WM_ENABLE", WM_ENABLE);
    MsgTbl->Add("WM_SETREDRAW", WM_SETREDRAW);
    MsgTbl->Add("WM_SETTEXT", WM_SETTEXT);
    MsgTbl->Add("WM_GETTEXT", WM_GETTEXT);
    MsgTbl->Add("WM_GETTEXTLENGTH", WM_GETTEXTLENGTH);
    MsgTbl->Add("WM_PAINT", WM_PAINT);
    MsgTbl->Add("WM_CLOSE", WM_CLOSE);
    MsgTbl->Add("WM_QUERYENDSESSION", WM_QUERYENDSESSION);
    MsgTbl->Add("WM_QUIT", WM_QUIT);
    MsgTbl->Add("WM_QUERYOPEN", WM_QUERYOPEN);
    MsgTbl->Add("WM_ERASEBKGND", WM_ERASEBKGND);
    MsgTbl->Add("WM_SYSCOLORCHANGE", WM_SYSCOLORCHANGE);
    MsgTbl->Add("WM_ENDSESSION", WM_ENDSESSION);
    MsgTbl->Add("WM_SHOWWINDOW", WM_SHOWWINDOW);
    MsgTbl->Add("WM_WININICHANGE", WM_WININICHANGE);
    MsgTbl->Add("WM_SETTINGCHANGE", WM_SETTINGCHANGE);
    MsgTbl->Add("WM_DEVMODECHANGE", WM_DEVMODECHANGE);
    MsgTbl->Add("WM_ACTIVATEAPP", WM_ACTIVATEAPP);
    MsgTbl->Add("WM_FONTCHANGE", WM_FONTCHANGE);
    MsgTbl->Add("WM_TIMECHANGE", WM_TIMECHANGE);
    MsgTbl->Add("WM_CANCELMODE", WM_CANCELMODE);
    MsgTbl->Add("WM_SETCURSOR", WM_SETCURSOR);
    MsgTbl->Add("WM_MOUSEACTIVATE", WM_MOUSEACTIVATE);
    MsgTbl->Add("WM_CHILDACTIVATE", WM_CHILDACTIVATE);
    MsgTbl->Add("WM_QUEUESYNC", WM_QUEUESYNC);
    MsgTbl->Add("WM_GETMINMAXINFO", WM_GETMINMAXINFO);
    MsgTbl->Add("WM_PAINTICON", WM_PAINTICON);
    MsgTbl->Add("WM_ICONERASEBKGND", WM_ICONERASEBKGND);
    MsgTbl->Add("WM_NEXTDLGCTL", WM_NEXTDLGCTL);
    MsgTbl->Add("WM_SPOOLERSTATUS", WM_SPOOLERSTATUS);
    MsgTbl->Add("WM_DRAWITEM", WM_DRAWITEM);
    MsgTbl->Add("WM_MEASUREITEM", WM_MEASUREITEM);
    MsgTbl->Add("WM_DELETEITEM", WM_DELETEITEM);
    MsgTbl->Add("WM_VKEYTOITEM", WM_VKEYTOITEM);
    MsgTbl->Add("WM_CHARTOITEM", WM_CHARTOITEM);
    MsgTbl->Add("WM_SETFONT", WM_SETFONT);
    MsgTbl->Add("WM_GETFONT", WM_GETFONT);
    MsgTbl->Add("WM_SETHOTKEY", WM_SETHOTKEY);
    MsgTbl->Add("WM_GETHOTKEY", WM_GETHOTKEY);
    MsgTbl->Add("WM_QUERYDRAGICON", WM_QUERYDRAGICON);
    MsgTbl->Add("WM_COMPAREITEM", WM_COMPAREITEM);
    MsgTbl->Add("WM_COMPACTING", WM_COMPACTING);
    MsgTbl->Add("WM_COMMNOTIFY", WM_COMMNOTIFY);
    MsgTbl->Add("WM_WINDOWPOSCHANGING", WM_WINDOWPOSCHANGING);
    MsgTbl->Add("WM_WINDOWPOSCHANGED", WM_WINDOWPOSCHANGED);
    MsgTbl->Add("WM_POWER", WM_POWER);
    MsgTbl->Add("WM_COPYDATA", WM_COPYDATA);
    MsgTbl->Add("WM_CANCELJOURNAL", WM_CANCELJOURNAL);
    MsgTbl->Add("WM_NOTIFY", WM_NOTIFY);
    MsgTbl->Add("WM_INPUTLANGCHANGEREQUEST", WM_INPUTLANGCHANGEREQUEST);
    MsgTbl->Add("WM_INPUTLANGCHANGE", WM_INPUTLANGCHANGE);
    MsgTbl->Add("WM_TCARD", WM_TCARD);
    MsgTbl->Add("WM_HELP", WM_HELP);
    MsgTbl->Add("WM_USERCHANGED", WM_USERCHANGED);
    MsgTbl->Add("WM_NOTIFYFORMAT", WM_NOTIFYFORMAT);
    MsgTbl->Add("WM_CONTEXTMENU", WM_CONTEXTMENU);
    MsgTbl->Add("WM_STYLECHANGING", WM_STYLECHANGING);
    MsgTbl->Add("WM_STYLECHANGED", WM_STYLECHANGED);
    MsgTbl->Add("WM_DISPLAYCHANGE", WM_DISPLAYCHANGE);
    MsgTbl->Add("WM_GETICON", WM_GETICON);
    MsgTbl->Add("WM_SETICON", WM_SETICON);
    MsgTbl->Add("WM_NCCREATE", WM_NCCREATE);
    MsgTbl->Add("WM_NCDESTROY", WM_NCDESTROY);
    MsgTbl->Add("WM_NCCALCSIZE", WM_NCCALCSIZE);
    MsgTbl->Add("WM_NCHITTEST", WM_NCHITTEST);
    MsgTbl->Add("WM_NCPAINT", WM_NCPAINT);
    MsgTbl->Add("WM_NCACTIVATE", WM_NCACTIVATE);
    MsgTbl->Add("WM_GETDLGCODE", WM_GETDLGCODE);
    MsgTbl->Add("WM_SYNCPAINT", WM_SYNCPAINT);
    MsgTbl->Add("WM_NCMOUSEMOVE", WM_NCMOUSEMOVE);
    MsgTbl->Add("WM_NCLBUTTONDOWN", WM_NCLBUTTONDOWN);
    MsgTbl->Add("WM_NCLBUTTONUP", WM_NCLBUTTONUP);
    MsgTbl->Add("WM_NCLBUTTONDBLCLK", WM_NCLBUTTONDBLCLK);
    MsgTbl->Add("WM_NCRBUTTONDOWN", WM_NCRBUTTONDOWN);
    MsgTbl->Add("WM_NCRBUTTONUP", WM_NCRBUTTONUP);
    MsgTbl->Add("WM_NCRBUTTONDBLCLK", WM_NCRBUTTONDBLCLK);
    MsgTbl->Add("WM_NCMBUTTONDOWN", WM_NCMBUTTONDOWN);
    MsgTbl->Add("WM_NCMBUTTONUP", WM_NCMBUTTONUP);
    MsgTbl->Add("WM_NCMBUTTONDBLCLK", WM_NCMBUTTONDBLCLK);
    MsgTbl->Add("WM_KEYFIRST", WM_KEYFIRST);
    MsgTbl->Add("WM_KEYDOWN", WM_KEYDOWN);
    MsgTbl->Add("WM_KEYUP", WM_KEYUP);
    MsgTbl->Add("WM_CHAR", WM_CHAR);
    MsgTbl->Add("WM_DEADCHAR", WM_DEADCHAR);
    MsgTbl->Add("WM_SYSKEYDOWN", WM_SYSKEYDOWN);
    MsgTbl->Add("WM_SYSKEYUP", WM_SYSKEYUP);
    MsgTbl->Add("WM_SYSCHAR", WM_SYSCHAR);
    MsgTbl->Add("WM_SYSDEADCHAR", WM_SYSDEADCHAR);
    MsgTbl->Add("WM_KEYLAST", WM_KEYLAST);
    MsgTbl->Add("WM_IME_STARTCOMPOSITION", WM_IME_STARTCOMPOSITION);
    MsgTbl->Add("WM_IME_ENDCOMPOSITION", WM_IME_ENDCOMPOSITION);
    MsgTbl->Add("WM_IME_COMPOSITION", WM_IME_COMPOSITION);
    MsgTbl->Add("WM_IME_KEYLAST", WM_IME_KEYLAST);
    MsgTbl->Add("WM_INITDIALOG", WM_INITDIALOG);
    MsgTbl->Add("WM_COMMAND", WM_COMMAND);
    MsgTbl->Add("WM_SYSCOMMAND", WM_SYSCOMMAND);
    MsgTbl->Add("WM_TIMER", WM_TIMER);
    MsgTbl->Add("WM_HSCROLL", WM_HSCROLL);
    MsgTbl->Add("WM_VSCROLL", WM_VSCROLL);
    MsgTbl->Add("WM_INITMENU", WM_INITMENU);
    MsgTbl->Add("WM_INITMENUPOPUP", WM_INITMENUPOPUP);
    MsgTbl->Add("WM_MENUSELECT", WM_MENUSELECT);
    MsgTbl->Add("WM_MENUCHAR", WM_MENUCHAR);
    MsgTbl->Add("WM_ENTERIDLE", WM_ENTERIDLE);
    MsgTbl->Add("WM_CTLCOLORMSGBOX", WM_CTLCOLORMSGBOX);
    MsgTbl->Add("WM_CTLCOLOREDIT", WM_CTLCOLOREDIT);
    MsgTbl->Add("WM_CTLCOLORLISTBOX", WM_CTLCOLORLISTBOX);
    MsgTbl->Add("WM_CTLCOLORBTN", WM_CTLCOLORBTN);
    MsgTbl->Add("WM_CTLCOLORDLG", WM_CTLCOLORDLG);
    MsgTbl->Add("WM_CTLCOLORSCROLLBAR", WM_CTLCOLORSCROLLBAR);
    MsgTbl->Add("WM_CTLCOLORSTATIC", WM_CTLCOLORSTATIC);
    MsgTbl->Add("WM_MOUSEFIRST", WM_MOUSEFIRST);
    MsgTbl->Add("WM_MOUSEMOVE", WM_MOUSEMOVE);
    MsgTbl->Add("WM_LBUTTONDOWN", WM_LBUTTONDOWN);
    MsgTbl->Add("WM_LBUTTONUP", WM_LBUTTONUP);
    MsgTbl->Add("WM_LBUTTONDBLCLK", WM_LBUTTONDBLCLK);
    MsgTbl->Add("WM_RBUTTONDOWN", WM_RBUTTONDOWN);
    MsgTbl->Add("WM_RBUTTONUP", WM_RBUTTONUP);
    MsgTbl->Add("WM_RBUTTONDBLCLK", WM_RBUTTONDBLCLK);
    MsgTbl->Add("WM_MBUTTONDOWN", WM_MBUTTONDOWN);
    MsgTbl->Add("WM_MBUTTONUP", WM_MBUTTONUP);
    MsgTbl->Add("WM_MBUTTONDBLCLK", WM_MBUTTONDBLCLK);
    MsgTbl->Add("WM_MOUSELAST", WM_MOUSELAST);
    MsgTbl->Add("WM_PARENTNOTIFY", WM_PARENTNOTIFY);
    MsgTbl->Add("WM_ENTERMENULOOP", WM_ENTERMENULOOP);
    MsgTbl->Add("WM_EXITMENULOOP", WM_EXITMENULOOP);
    MsgTbl->Add("WM_NEXTMENU", WM_NEXTMENU);
    MsgTbl->Add("WM_SIZING", WM_SIZING);
    MsgTbl->Add("WM_CAPTURECHANGED", WM_CAPTURECHANGED);
    MsgTbl->Add("WM_MOVING", WM_MOVING);
    MsgTbl->Add("WM_POWERBROADCAST", WM_POWERBROADCAST);
    MsgTbl->Add("WM_DEVICECHANGE", WM_DEVICECHANGE);
    MsgTbl->Add("WM_MDICREATE", WM_MDICREATE);
    MsgTbl->Add("WM_MDIDESTROY", WM_MDIDESTROY);
    MsgTbl->Add("WM_MDIACTIVATE", WM_MDIACTIVATE);
    MsgTbl->Add("WM_MDIRESTORE", WM_MDIRESTORE);
    MsgTbl->Add("WM_MDINEXT", WM_MDINEXT);
    MsgTbl->Add("WM_MDIMAXIMIZE", WM_MDIMAXIMIZE);
    MsgTbl->Add("WM_MDITILE", WM_MDITILE);
    MsgTbl->Add("WM_MDICASCADE", WM_MDICASCADE);
    MsgTbl->Add("WM_MDIICONARRANGE", WM_MDIICONARRANGE);
    MsgTbl->Add("WM_MDIGETACTIVE", WM_MDIGETACTIVE);
    MsgTbl->Add("WM_MDISETMENU", WM_MDISETMENU);
    MsgTbl->Add("WM_ENTERSIZEMOVE", WM_ENTERSIZEMOVE);
    MsgTbl->Add("WM_EXITSIZEMOVE", WM_EXITSIZEMOVE);
    MsgTbl->Add("WM_DROPFILES", WM_DROPFILES);
    MsgTbl->Add("WM_MDIREFRESHMENU", WM_MDIREFRESHMENU);
    MsgTbl->Add("WM_IME_SETCONTEXT", WM_IME_SETCONTEXT);
    MsgTbl->Add("WM_IME_NOTIFY", WM_IME_NOTIFY);
    MsgTbl->Add("WM_IME_CONTROL", WM_IME_CONTROL);
    MsgTbl->Add("WM_IME_COMPOSITIONFULL", WM_IME_COMPOSITIONFULL);
    MsgTbl->Add("WM_IME_SELECT", WM_IME_SELECT);
    MsgTbl->Add("WM_IME_CHAR", WM_IME_CHAR);
    MsgTbl->Add("WM_IME_KEYDOWN", WM_IME_KEYDOWN);
    MsgTbl->Add("WM_IME_KEYUP", WM_IME_KEYUP);
    MsgTbl->Add("WM_CUT", WM_CUT);
    MsgTbl->Add("WM_COPY", WM_COPY);
    MsgTbl->Add("WM_PASTE", WM_PASTE);
    MsgTbl->Add("WM_CLEAR", WM_CLEAR);
    MsgTbl->Add("WM_UNDO", WM_UNDO);
    MsgTbl->Add("WM_RENDERFORMAT", WM_RENDERFORMAT);
    MsgTbl->Add("WM_RENDERALLFORMATS", WM_RENDERALLFORMATS);
    MsgTbl->Add("WM_DESTROYCLIPBOARD", WM_DESTROYCLIPBOARD);
    MsgTbl->Add("WM_DRAWCLIPBOARD", WM_DRAWCLIPBOARD);
    MsgTbl->Add("WM_PAINTCLIPBOARD", WM_PAINTCLIPBOARD);
    MsgTbl->Add("WM_VSCROLLCLIPBOARD", WM_VSCROLLCLIPBOARD);
    MsgTbl->Add("WM_SIZECLIPBOARD", WM_SIZECLIPBOARD);
    MsgTbl->Add("WM_ASKCBFORMATNAME", WM_ASKCBFORMATNAME);
    MsgTbl->Add("WM_CHANGECBCHAIN", WM_CHANGECBCHAIN);
    MsgTbl->Add("WM_HSCROLLCLIPBOARD", WM_HSCROLLCLIPBOARD);
    MsgTbl->Add("WM_QUERYNEWPALETTE", WM_QUERYNEWPALETTE);
    MsgTbl->Add("WM_PALETTEISCHANGING", WM_PALETTEISCHANGING);
    MsgTbl->Add("WM_PALETTECHANGED", WM_PALETTECHANGED);
    MsgTbl->Add("WM_HOTKEY", WM_HOTKEY);
    MsgTbl->Add("WM_PRINT", WM_PRINT);
    MsgTbl->Add("WM_PRINTCLIENT", WM_PRINTCLIENT);
    MsgTbl->Add("WM_HANDHELDFIRST", WM_HANDHELDFIRST);
    MsgTbl->Add("WM_HANDHELDLAST", WM_HANDHELDLAST);
    MsgTbl->Add("WM_AFXFIRST", WM_AFXFIRST);
    MsgTbl->Add("WM_AFXLAST", WM_AFXLAST);
    MsgTbl->Add("WM_PENWINFIRST", WM_PENWINFIRST);
    MsgTbl->Add("WM_PENWINLAST", WM_PENWINLAST);
    MsgTbl->Add("WM_APP", WM_APP);
    MsgTbl->Add("WM_USER", WM_USER);
}


extern "C" int
SendMsg_Init(Tcl_Interp *interp)
{
    Tcl_MutexLock(&initLock);

    if (MsgTbl == 0L) {
	Fill_MsgTable();
    }

    Tcl_MutexUnlock(&initLock);

    Tcl_CreateObjCommand(interp, "winutils::sendmessage", SendMessageObjCmd, 0L, 0L);
    return TCL_OK;
}
