/* ----------------------------------------------------------------------------
 * Volume.c --
 *
 *	The volinfo command.  Returns info about volumes (disks).
 *
 * ----------------------------------------------------------------------------
 *
 * Written by: David Gravereaux <davygrvy@pobox.com>
 *
 * No restrictions are placed on the redistribution of this.  See license.txt
 * for more detail.
 *
 * ----------------------------------------------------------------------------
 * RCS: @(#) $Id: Volume.c,v 1.3 2002/04/21 06:19:37 davygrvy Exp $
 * ----------------------------------------------------------------------------
 */

#include "tcl.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

static Tcl_Obj *GetSysMsgObj(DWORD id)
{
    int chars;
    char sysMsgSpace[512];

    chars = wsprintf(sysMsgSpace, "[%u] ", id);
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
	    FORMAT_MESSAGE_IGNORE_INSERTS|FORMAT_MESSAGE_MAX_WIDTH_MASK,
	    0L, id, 0, sysMsgSpace+chars, (512-chars), 0);
    return Tcl_NewStringObj(sysMsgSpace, -1);
}


static int
GetVolTypeObjCmd (ClientData cdata, Tcl_Interp *interp,
		  int objc, Tcl_Obj *CONST objv[])
{
    Tcl_Obj *result = Tcl_GetObjResult(interp);
    TCHAR volName[MAX_PATH];
    DWORD serial;
    DWORD maxFileNameCompLen;
    DWORD fsFlags;
    TCHAR fsName[MAX_PATH];
    UINT oldMode;
    int rtn;

    if (objc < 2) {
	/* no drives to query! */
	Tcl_WrongNumArgs(interp, 1, objv, "<drive>");
	return TCL_ERROR;
    }

    oldMode = SetErrorMode(SEM_FAILCRITICALERRORS);

    if (!GetVolumeInformation(Tcl_GetString(objv[1]), volName, MAX_PATH,
	    &serial, &maxFileNameCompLen, &fsFlags, fsName, MAX_PATH))
    {
	Tcl_SetObjResult(interp, GetSysMsgObj(GetLastError()));
	rtn = TCL_ERROR;
    } else {
	Tcl_ListObjAppendElement(0, result, Tcl_NewStringObj(volName, -1));
	Tcl_ListObjAppendElement(0, result, Tcl_NewLongObj(serial));
	Tcl_ListObjAppendElement(0, result, Tcl_NewStringObj(fsName, -1));
	rtn = TCL_OK;
    }

    SetErrorMode(oldMode);
    return rtn;
}

int
Vol_Init(Tcl_Interp *interp)
{
    Tcl_CreateObjCommand(interp, "winutils::volinfo", GetVolTypeObjCmd, 0L, 0L);
    return TCL_OK;
}