/*************************************************************
 ** winutils.c  --
 **
 **	Interp Init.
 **
 ** Author: David Gravereaux  mailto:davygrvy@pobox.com
 ************************************************************/

#include "winutils.h"
#include <windows.h>

#ifdef _MSC_VER
    /* Only do this when MSVC++ is compiling us. */
#   ifdef USE_TCL_STUBS
	/* Mark this .obj as needing tcl's Stubs library. */
#	pragma comment(lib, "tclstub" \
		STRINGIFY(JOIN(TCL_MAJOR_VERSION,TCL_MINOR_VERSION)) ".lib")
#	if !defined(_MT) || !defined(_DLL) || defined(_DEBUG)
	    /* This fixes a bug with how the Stubs library was compiled.
	     * The requirement for msvcrt.lib from tclstubXX.lib should
	     * be removed. */
#	    pragma comment(linker, "-nodefaultlib:msvcrt.lib")
#	endif
#   else
	/* Mark this .obj needing the import library */
#	pragma comment(lib, "tcl" \
		STRINGIFY(JOIN(TCL_MAJOR_VERSION,TCL_MINOR_VERSION)) ".lib")
#   endif
#endif

#undef TCL_STORAGE_CLASS
#define TCL_STORAGE_CLASS DLLEXPORT

EXTERN int
Winutils_Init (Tcl_Interp *interp)
{
#ifdef USE_TCL_STUBS
    if (Tcl_InitStubs(interp, "8.1", 0) == 0L) {
	return TCL_ERROR;
    }
#endif

/*    if (Dialogbox_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }*/
    if (Dirwatch_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Hotkey_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Keylock_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Launch_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Perfmon_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (ProcessExtras_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Reg9XService_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (SendMsg_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Shell_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Uptime_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Vol_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Winhelp_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Winsockmisc_Init(interp) != TCL_OK) {
	return TCL_ERROR;
    }

    Tcl_PkgProvide(interp, "winutils", "0.8");
    return TCL_OK;
}

EXTERN BOOL WINAPI
DllMain (HINSTANCE hInst, ULONG ulReason, LPVOID lpReserved)
{
    if (ulReason == DLL_PROCESS_ATTACH) {
	DisableThreadLibraryCalls(hInst);
    }
    return TRUE; /* successful load */
}